import abjad
import auxjad


def test_LeafDynMaker_01():
    pitches = [0, 2, 4, 5, 7, 9]
    durations = [(1, 32), (2, 32), (3, 32), (4, 32), (5, 32), (6, 32)]
    dynamics = ['pp', 'p', 'mp', 'mf', 'f', 'ff']
    articulations = ['.', '>', '-', '_', '^', '+']
    leaf_dyn_maker = auxjad.LeafDynMaker()
    notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'32
            \pp
            - \staccato
            d'16
            \p
            - \accent
            e'16.
            \mp
            - \tenuto
            f'8
            \mf
            - \portato
            g'8
            \f
            - \marcato
            ~
            g'32
            a'8.
            \ff
            - \stopped
        }
        ''')


def test_LeafDynMaker_02():
    pitches = [5, None, (0, 2, 7)]
    durations = [(1, 4), (1, 8), (1, 16)]
    dynamics = ['p', None, 'f']
    articulations = ['staccato', None, 'tenuto']
    leaf_dyn_maker = auxjad.LeafDynMaker()
    notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            f'4
            \p
            - \staccato
            r8
            <c' d' g'>16
            \f
            - \tenuto
        }
        ''')


def test_LeafDynMaker_03():
    pitches = [0, 2, 4, 5, 7, 9]
    durations = [(1, 32), (2, 32), (3, 32), (4, 32), (5, 32), (6, 32)]
    dynamics = ['pp', 'pp', 'mp', 'f', 'f', 'p']
    leaf_dyn_maker = auxjad.LeafDynMaker()
    notes = leaf_dyn_maker(pitches,
                           durations,
                           dynamics,
                           no_repeat=True,
                           )
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'32
            \pp
            d'16
            e'16.
            \mp
            f'8
            \f
            g'8
            ~
            g'32
            a'8.
            \p
        }
        ''')


def test_LeafDynMaker_04():
    pitches = [0, 2, 4, 5, 7, 9]
    durations = (1, 4)
    dynamics = ['p', 'f', 'ff']
    articulations = ['.', '>']
    leaf_dyn_maker = auxjad.LeafDynMaker()
    notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \p
            - \staccato
            d'4
            \f
            - \accent
            e'4
            \ff
            f'4
            g'4
            a'4
        }
        ''')


def test_LeafDynMaker_05():
    pitches = [0, 2, 4, 5, 7, 9]
    durations = (1, 4)
    dynamics = 'p'
    articulations = '.'
    leaf_dyn_maker = auxjad.LeafDynMaker()
    notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \p
            - \staccato
            d'4
            - \staccato
            e'4
            - \staccato
            f'4
            - \staccato
            g'4
            - \staccato
            a'4
            - \staccato
        }
        ''')


def test_LeafDynMaker_06():
    pitches = [0,
               "d'",
               'E4',
               abjad.NumberedPitch(5),
               abjad.NamedPitch("g'"),
               abjad.NamedPitch("A4"),
               ]
    durations = [(1, 32),
                 "2/32",
                 abjad.Duration("3/32"),
                 abjad.Duration(0.125),
                 abjad.Duration(5, 32),
                 abjad.Duration(6/32),
                 ]
    leaf_dyn_maker = auxjad.LeafDynMaker()
    notes = leaf_dyn_maker(pitches, durations)
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'32
            d'16
            e'16.
            f'8
            g'8
            ~
            g'32
            a'8.
        }
        ''')



import abjad


class LeafDynMaker(abjad.LeafMaker):
    r"""An extension of ``abjad.LeafMaker`` which can also take optional lists
    of dynamics and articulations.

    ..  container:: example

        Usage is similar to ``abjad.LeafMaker``:

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = [(1, 32), (2, 32), (3, 32), (4, 32), (5, 32), (6, 32)]
        >>> dynamics = ['pp', 'p', 'mp', 'mf', 'f', 'ff']
        >>> articulations = ['.', '>', '-', '_', '^', '+']
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'32
            \pp
            -\staccato
            d'16
            \p
            -\accent
            e'16.
            \mp
            -\tenuto
            f'8
            \mf
            -\portato
            g'8
            \f
            -\marcato
            ~
            g'32
            a'8.
            \ff
            -\stopped
        }

    ..  container:: example

        Tuple elements in ``pitches`` result in chords. None-valued elements
        in ``pitches`` result in rests:

        >>> pitches = [5, None, (0, 2, 7)]
        >>> durations = [(1, 4), (1, 8), (1, 16)]
        >>> dynamics = ['p', None, 'f']
        >>> articulations = ['staccato', None, 'tenuto']
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            f'4
            \p
            -\staccato
            r8
            <c' d' g'>16
            \f
            -\tenuto
        }

    ..  container:: example

        Can omit repeated dynamics with the keyword argument ``no_repeat``:

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = [(1, 32), (2, 32), (3, 32), (4, 32), (5, 32), (6, 32)]
        >>> dynamics = ['pp', 'pp', 'mp', 'f', 'f', 'p']
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches,
        ...                        durations,
        ...                        dynamics,
        ...                        no_repeat=True,
        ...                        )
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'32
            \pp
            d'16
            e'16.
            \mp
            f'8
            \f
            g'8
            ~
            g'32
            a'8.
            \p
        }

    ..  container:: example

        The lengths of both ``dynamics`` and ``articulations`` can be shorter
        than the lengths of ``pitches`` and ``durations`` (whatever is the
        greatest):

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = (1, 4)
        >>> dynamics = ['p', 'f', 'ff']
        >>> articulations = ['.', '>']
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \p
            -\staccato
            d'4
            \f
            -\accent
            e'4
            \ff
            f'4
            g'4
            a'4
        }

    ..  container:: example

        If the length of ``articulations`` is 1, it will apply to all elements.
        If the length of ``dynamics`` is 1, it will apply to the first element
        only:

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = (1, 4)
        >>> dynamics = 'p'
        >>> articulations = '.'
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \p
            -\staccato
            d'4
            -\staccato
            e'4
            -\staccato
            f'4
            -\staccato
            g'4
            -\staccato
            a'4
            -\staccato
        }

    ..  container:: example

        Similarly to Abjad's native classes, it accepts many types of elements
        in its input lists:

        >>> pitches = [0,
        ...            "d'",
        ...            'E4',
        ...            abjad.NumberedPitch(5),
        ...            abjad.NamedPitch("g'"),
        ...            abjad.NamedPitch("A4"),
        ...            ]
        >>> durations = [(1, 32),
        ...              "2/32",
        ...              abjad.Duration("3/32"),
        ...              abjad.Duration(0.125),
        ...              abjad.Duration(5, 32),
        ...              abjad.Duration(6/32),
        ...              ]
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches, durations)
        >>> staff = abjad.Staff(notes)
        \new Staff
        {
            c'32
            d'16
            e'16.
            f'8
            g'8
            ~
            g'32
            a'8.
        }
    """

    def __call__(self,
                 pitches,
                 durations,
                 dynamics=None,
                 articulations=None,
                 *,
                 no_repeat=False
                 ) -> abjad.Selection:
        if dynamics:
            for dynamic in dynamics:
                if dynamic is not None:
                    if not isinstance(dynamic, (str, abjad.Dynamic)):
                        raise TypeError("'dynamic' must be 'str' or "
                                        "'abjad.Dynamic'")
        if articulations:
            for articulation in articulations:
                if articulation is not None:
                    if not isinstance(articulation, (str, abjad.Articulation)):
                        raise TypeError("'articulation' must be 'str' or "
                                        "'abjad.Articulation'")

        leaves = super().__call__(pitches, durations)
        dummy_container = abjad.Container(leaves)
        logical_ties = leaves.logical_ties()

        dynamics_ = self._listify(dynamics)
        articulations_ = self._listify(articulations)
        greatest_len = max(len(pitches), len(durations))
        self._fill_list(dynamics_, greatest_len)
        self._fill_list(articulations_,
                        greatest_len,
                        default=self._single_element_or_none(articulations_),
                        )

        previous_dynamic = None
        for logical_tie, dynamic, articulation in zip(logical_ties,
                                                      dynamics_,
                                                      articulations_):
            if (not no_repeat or dynamic != previous_dynamic) and dynamic:
                abjad.attach(abjad.Dynamic(dynamic), logical_tie.head)
                previous_dynamic = dynamic
            if articulation:
                abjad.attach(abjad.Articulation(articulation),
                             logical_tie.head,
                             )

        result = dummy_container[:]
        dummy_container[:] = []

        return result

    @staticmethod
    def _listify(argument):
        if argument:
            if isinstance(argument, list):
                return argument
            else:
                return [argument]
        else:
            return []

    @staticmethod
    def _single_element_or_none(input_list):
        if len(input_list) == 1:
            return input_list[0]
        else:
            return None

    @staticmethod
    def _fill_list(input_list, length, default=None):
        while len(input_list) < length:
            input_list.append(default)



from ._LoopWindowGeneric import _LoopWindowGeneric


class LoopWindowByList(_LoopWindowGeneric):
    r"""Similar to ``LoopWindowByList``, but instead of taking an
    ``abjad.Container`` input, it takes a list of arbitrary size. It then
    outputs the list elements, whatever they may be. The list elements can be
    ``abjad.Container``'s, but they can also be anything else, thus being more
    general. Takes a list as input as well as an integer representing the
    number of elements per looping window, then outputs individual elements
    with according to the looping process. For instance, if the initial list
    had the elements ``[A, B, C, D, E, F]`` and the looping window was size
    three, the output would be:

    ``A B C B C D C D E D E F E F F``

    This can be better visualised as:

    .. code-block:: none

        A B C
          B C D
            C D E
              D E F
                E F
                  F

    ..  container:: example

        It takes a list and the number of elements of the window as arguments.
        Each call of the object, in this case ``looper()``, will move the
        window forwards and output the result:

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopWindowByList(input_list, window_size=3)
        >>> looper()
        ['A', 'B', 'C']
        >>> looper()
        ['B', 'C', 'D']

        The method get_current_window() will output the current window without
        moving the head forwards.

        >>> looper.get_current_window()
        ['B', 'C', 'D']

    ..  container:: example

        The instances of ``LoopWindowByList`` can also be used as an iterator,
        which can then be used in a for loop to exhaust all windows.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopWindowByList(input_list,
        ...                                  window_size=3,
        ...                                  )
        >>> for window in looper:
        ...     print(window)
        ['A', 'B', 'C']
        ['B', 'C', 'D']
        ['C', 'D', 'E']
        ['D', 'E', 'F']
        ['E', 'F']
        ['F']

    ..  container:: example

        This class can take many optional keyword arguments during its
        creation. ``step_size`` dictates the size of each individual step in
        number of elements (default value is 1). ``max_steps`` sets the maximum
        number of steps that the window can advance when the object is called,
        ranging between 1 and the input value (default is also 1).
        ``repetition_chance`` sets the chance of a window result repeating
        itself (that is, the window not moving forwards when called). It should
        range from 0.0 to 1.0 (default 0.0, i.e. no repetition). 
        ``forward_bias`` sets the chance of the window moving forward instead
        of backwards. It should range from 0.0 to 1.0 (default 1.0, which means
        the window can only move forwards. A value of 0.5 gives 50% chance of
        moving forwards while a value of 0.0 will move the window only
        backwards). Finally, ``head_position`` can be used to offset the
        starting position of the looping window (default is 0).

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopWindowByList(input_list,
        ...                                  window_size=3,
        ...                                  step_size=1,
        ...                                  max_steps=2,
        ...                                  repetition_chance=0.25,
        ...                                  forward_bias=0.2,
        ...                                  head_position=0,
        ...                                  )
        >>> looper.window_size
        3
        >>> looper.step_size
        1
        >>> looper.repetition_chance
        0.25
        >>> looper.forward_bias
        0.2
        >>> looper.max_steps
        2
        >>> looper.head_position
        0

        Use the ``set_`` methods below to change these values after
        initialisation.

        >>> looper.set_window_size(2)
        >>> looper.set_step_size(2)
        >>> looper.set_max_steps(3)
        >>> looper.set_repetition_chance(0.1)
        >>> looper.set_forward_bias(0.8)
        >>> looper.set_head_position(2)
        >>> looper.window_size
        2
        >>> looper.step_size
        2
        >>> looper.max_steps
        3
        >>> looper.repetition_chance
        0.1
        >>> looper.forward_bias
        0.8
        >>> looper.head_position
        2

    ..  container:: example

        The function ``len()`` can be used to get the total number of elements
        in the container.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopWindowByList(input_list, window_size=3)
        >>> len(looper)
        6

    ..  container:: example

        To run through just part of the process and output it as a single
        list, starting from the initial head position, use the method
        ``output_n()`` and pass the number of iterations as argument.

        >>> input_list = ['A', 'B', 'C', 'D']
        >>> looper = auxjad.LoopWindowByList(input_list, window_size=3)
        >>> looper.output_n(2)
        ['A', 'B', 'C', 'B', 'C', 'D']

    ..  container:: example

        To run through the whole process and output it as a single list, from
        the initial head position until the process outputs the single last
        element, use the method ``output_all()``.

        >>> input_list = ['A', 'B', 'C', 'D']
        >>> looper = auxjad.LoopWindowByList(input_list, window_size=3)
        >>> looper.output_all()
        ['A', 'B', 'C', 'B', 'C', 'D', 'C', 'D', 'D']

    .. container:: example

        To change the size of the window after instantiation, use the method
        ``set_window_size()``. In the example below, the initial window is of
        size 3, and so the first call of the looper object outputs the first,
        second, and third elements of the list. The window size is then set to
        4, and the looper is called again, moving to the element in the next
        position, thus outputting the second, third, fourth, and fifth
        elements.

        >>> input_list = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> looper = auxjad.LoopWindowByList(input_list, window_size=3)
        >>> looper()
        ['A', 'B', 'C']
        >>> looper.set_window_size(4)
        >>> looper()
        ['B', 'C', 'D', 'E']

    .. container:: example

        It should be clear that the list can contain any types of elements:

        >>> input_list = [123, 'foo', (3, 4), 3.14]
        >>> looper = auxjad.LoopWindowByList(input_list, window_size=3)
        >>> looper()
        [123, 'foo', (3, 4)]

        This also include Abjad's types. Abjad's exclusive membership
        requirement is respected since each call returns a ``copy.deepcopy`` of
        the window. The same is true to the ``output_all()`` method.

        >>> import abjad
        >>> import copy
        >>> input_list = [
        ...     abjad.Container(r"c'4 d'4 e'4 f'4"),
        ...     abjad.Container(r"fs'1"),
        ...     abjad.Container(r"r2 bf2"),
        ...     abjad.Container(r"c''2. r4"),
        ... ]
        >>> looper = auxjad.LoopWindowByList(input_list, window_size=3)
        >>> staff = abjad.Staff()
        >>> for element in looper.output_all():
        ...     staff.append(element)
        >>> abjad.f(staff)
        \new Staff
        {
            {
                c'4
                d'4
                e'4
                f'4
            }
            {
                fs'1
            }
            {
                r2
                bf2
            }
            {
                fs'1
            }
            {
                r2
                bf2
            }
            {
                c''2.
                r4
            }
            {
                r2
                bf2
            }
            {
                c''2.
                r4
            }
            {
                c''2.
                r4
            }
        }
    """

    def __init__(self,
                 container: list,
                 *,
                 window_size: int,
                 step_size: int = 1,
                 max_steps: int = 1,
                 repetition_chance: float = 0.0,
                 forward_bias: float = 1.0,
                 head_position: int = 0,
                 ):
        if not isinstance(container, list):
            raise TypeError("'container' must be 'list'")
        self._container = container[:]
        super().__init__(head_position,
                         window_size,
                         step_size,
                         max_steps,
                         repetition_chance,
                         forward_bias,
                         )

    def __len__(self) -> int:
        return len(self._container)

    def __next__(self) -> list:
        r"""Custom __next__ dunder method since parent's method outputs an
        ``abjad.Selection``.
        """
        self._move_head()
        if self._done():
            raise StopIteration
        self._slice_container()
        return self._current_window[:]

    def output_all(self) -> list:
        r"""Custom output_all() method since parent's method outputs an
        ``abjad.Selection``.
        """
        dummy_container = []
        while True:
            try:
                dummy_container.extend(self.__call__())
            except:
                break
        return dummy_container[:]

    def output_n(self, n: int) -> list:
        r"""Custom output_n() method since parent's method outputs an
        ``abjad.Selection``.
        """
        dummy_container = []
        for _ in range(n):
            dummy_container.extend(self.__call__())
        return dummy_container[:]

    def _slice_container(self) -> list:
        start = self.head_position
        end = self.head_position + self.window_size
        self._current_window = self._container[start:end]



import abjad


def underfull_duration(container: abjad.Container) -> abjad.Duration:
    r"""Checks if an ``abjad.Container`` is underfull and returns the missing
    ``abjad.Duration``.

    ..  container:: example

        Returns the missing duration of the last bar of any container or child
        class. If no time signature is encountered, it uses LilyPond's
        convention and considers the container as in 4/4.

        >>> container1 = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"c'4 d'4 e'4")
        >>> container3 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4")
        >>> container4 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4 d'4 e'4 f'4")
        >>> auxjad.underfull_duration(container1)
        0
        >>> auxjad.underfull_duration(container2)
        1/4
        >>> auxjad.underfull_duration(container3)
        3/4
        >>> auxjad.underfull_duration(container4)
        0

    ..  container:: example

        Handles any time signatures as well as changes of time signature.

        >>> container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"\time 3/4 a2. \time 2/4 r2")
        >>> container3 = abjad.Container(r"\time 5/4 g1 ~ g4 \time 4/4 af'2")
        >>> container4 = abjad.Container(r"\time 6/8 c'2 ~ c'8")
        >>> auxjad.underfull_duration(container1)
        0
        >>> auxjad.underfull_duration(container2)
        0
        >>> auxjad.underfull_duration(container3)
        1/2
        >>> auxjad.underfull_duration(container4)
        1/8

    ..  container:: example

        Correctly handles partial time signatures.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> time_signature = abjad.TimeSignature((3, 4), partial=(1, 4))
        >>> abjad.attach(time_signature, container[0])
        >>> auxjad.underfull_duration(container)
        0

    ..  container:: example

        It also handles multi-measure rests.

        >>> container1 = abjad.Container(r"R1")
        >>> container2 = abjad.Container(r"\time 3/4 R1*3/4 \time 2/4 r2")
        >>> container3 = abjad.Container(r"\time 5/4 R1*5/4 \time 4/4 g''4")
        >>> container4 = abjad.Container(r"\time 6/8 R1*1/2")
        >>> auxjad.underfull_duration(container1)
        0
        >>> auxjad.underfull_duration(container2)
        0
        >>> auxjad.underfull_duration(container3)
        3/4
        >>> auxjad.underfull_duration(container4)
        1/4

    ..  container:: example

        If a container is malformed, i.e. it has an underfilled bar before a
        time signature change, the function raises a ``ValueError`` exception.

        >>> container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'1")
        >>> auxjad.underfull_duration(container)
        ValueError: 'container' is malformed, with an underfull bar preceeding
        a time signature change
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("'container' must be 'abjad.Container' or child class")
    leaves = abjad.select(container).leaves()
    # handling first leaf
    leaf = leaves[0]
    time_signature = abjad.inspect(leaf).effective(abjad.TimeSignature)
    if time_signature:
        effective_time_signature = time_signature
    else:
        effective_time_signature = abjad.TimeSignature((4, 4))
    duration = abjad.inspect(leaf).duration()
    # handling partial time signatures
    if effective_time_signature.partial:
        duration += effective_time_signature.duration \
                  - effective_time_signature.partial
    # all other leaves
    for leaf in leaves[1:]:
        time_signature = abjad.inspect(leaf).effective(abjad.TimeSignature)
        if time_signature and time_signature != effective_time_signature:
            if duration % effective_time_signature.duration != 0:
                raise ValueError("'container' is malformed, with an underfull "
                                 "bar preceeding a time signature change")
            effective_time_signature = time_signature
            duration = abjad.Duration(0)
        duration += abjad.inspect(leaf).duration()
    duration_last_bar = duration % effective_time_signature.duration
    duration_left = duration_last_bar
    if duration_last_bar > abjad.Duration(0):
        duration_left = effective_time_signature.duration - duration_last_bar
    return duration_left



import abjad


def are_containers_equal(container1: abjad.Container,
                         container2: abjad.Container,
                         *,
                         include_indicators: bool = False,
                         ) -> bool:
    r"""A comparator function returning ``True`` when two containers are
    identical and ``False`` when they are not.

    ..  container:: example

        When the pitches and effective durations of all leaves in both
        containers are identical, this function returns ``True``:

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> auxjad.are_containers_equal(container1, container2)
        True

    ..  container:: example

        Even if all leaves of both containers are identical in relation to both
        pitches and written durations, the function considers the effective
        durations. This means that situations like the one below do not yield a
        false positive:

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(r"\times 3/2 {c'4 d'4 e'4} "
        ...                          "f'4 <g' a'>2 r2")
        >>> auxjad.are_containers_equal(container1, container2)
        False

    ..  container:: example

        By default, this function ignores indicators, so the containers in the
        example below are understood to be identical:

        >>> container1 = abjad.Staff(r"c'4\pp d'4 e'4-. f'4 <g' a'>2-> r2")
        >>> container2 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> auxjad.are_containers_equal(container1, container2)
        True

    ..  container:: example

        Setting the argument ``include_indicators`` to ``True`` forces the
        function to include indicators in its comparison. In that case, the
        containers in the example above are not considered identical any
        longer:

        >>> container1 = abjad.Staff(r"c'4\pp d'4 e'4-. f'4 <g' a'>2-> r2")
        >>> container2 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> auxjad.are_containers_equal(container1,
        ...                             container2,
        ...                             include_indicators=True,
        ...                             )
        True

    ..  container:: example

        This function also handles grace notes:

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Staff(r"c'4 \grace{d'4} d'4 e'4 f'4")
        >>> auxjad.are_containers_equal(container1, container2)
        False

        >>> container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(r"c'4 \grace{c''4} d'4 e'4 "
        ...                          "f'4 <g' a'>2 r2")
        >>> auxjad.are_containers_equal(container1, container2)
        False

        >>> container1 = abjad.Staff(r"c'4 \grace{c''4} d'4 e'4 "
        ...                          "f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(r"c'4 \grace{c''8} d'4 e'4 "
        ...                          "f'4 <g' a'>2 r2")
        >>> auxjad.are_containers_equal(container1, container2)
        False

        >>> container1 = abjad.Staff(r"c'4 \grace{c''16} d'4 e'4 "
        ...                          "f'4 <g' a'>2 r2")
        >>> container2 = abjad.Staff(r"c'4 \grace{c''16} d'4 e'4 "
        ...                          "f'4 <g' a'>2 r2")
        >>> auxjad.are_containers_equal(container1, container2)
        True
    """
    if not isinstance(container1, abjad.Container):
        raise TypeError("'container1' must be 'abjad.Container' or "
                        "child class")
    if not isinstance(container2, abjad.Container):
        raise TypeError("'container2' must be 'abjad.Container' or "
                        "child class")
    if not isinstance(include_indicators, bool):
        raise TypeError("'include_indicators' must be 'bool'")

    leaves1 = [leaf for leaf in abjad.select(container1).leaves()]
    leaves2 = [leaf for leaf in abjad.select(container2).leaves()]

    if len(leaves1) != len(leaves2):
        return False

    for leaf1, leaf2 in zip(leaves1, leaves2):
        if type(leaf1) != type(leaf2):
            return False
        if abjad.inspect(leaf1).duration() != abjad.inspect(leaf2).duration():
            return False
        if type(leaf1) == abjad.Note:
            if leaf1.written_pitch != leaf2.written_pitch:
                return False
        if type(leaf1) == abjad.Chord:
            if leaf1.written_pitches != leaf2.written_pitches:
                return False
        if type(abjad.inspect(leaf1).before_grace_container()) != \
                type(abjad.inspect(leaf2).before_grace_container()):
            return False
        if include_indicators:
            if abjad.inspect(leaf1).indicators() != \
                    abjad.inspect(leaf2).indicators():
                return False
    return True



try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

import auxjad

auxjad_classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3 :: Only',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Utilities',
]

with open('README.rst', 'r') as file:
    auxjad_long_description = file.read()

setup(
    name='auxjad',
    description='Auxiliary functions and classes for Abjad 3.1.',
    author='Gilberto Agostinho',
    author_email='gilbertohasnofb@gmail.com',
    version=auxjad.__version__,
    packages=find_packages(),
    url='https://github.com/gilbertohasnofb/auxjad',
    license='MIT',
    long_description=auxjad_long_description,
    tests_require=['abjad==3.1', 'pytest'],
    classifiers=auxjad_classifiers,
    python_requires='>=3.6',
    install_requires=['abjad==3.1', 'setuptools'],
)



import abjad


def remove_repeated_time_signatures(container: abjad.Container
                                    ) -> abjad.Container:
    r"""A function which removes all unecessary time signatures. It removes
    consecutive effective time signatures, even if separated by any number of
    bars with no time signature.

    ..  container:: example

        When two consecutive bars have identical time signatures, the second
        one is removed:

        >>> staff = abjad.Staff(r"c'4 d'8 | c'4 d'8")
        >>> abjad.attach(abjad.TimeSignature((3, 8)), staff[0])
        >>> abjad.attach(abjad.TimeSignature((3, 8)), staff[2])
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            \time 3/8
            c'4
            d'8
        }
        >>> staff = auxjad.remove_repeated_time_signatures(staff)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            c'4
            d'8
        }

    ..  container:: example

        The function also removes time signatures that are separated by an
        arbitrary number of bars without one:

        >>> staff = abjad.Staff(r"c'4 d'8 e'4. c'4 d'8")
        >>> abjad.attach(abjad.TimeSignature((3, 8)), staff[0])
        >>> abjad.attach(abjad.TimeSignature((3, 8)), staff[3])
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            e'4.
            \time 3/8
            c'4
            d'8
        }
        >>> staff = auxjad.remove_repeated_time_signatures(staff)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            e'4.
            c'4
            d'8
        }

    ..  container:: example

        The input container can also handle subcontainers, including cases in
        which the time signatures are attached to leaves of subcontainers:

        >>> staff = abjad.Staff([abjad.Note("c'2"),
        ...                      abjad.Chord("<d' f'>2"),
        ...                      abjad.Tuplet((2, 3), "g2 a2 b2"),
        ...                      ])
        >>> abjad.attach(abjad.TimeSignature((2, 2)), staff[0])
        >>> abjad.attach(abjad.TimeSignature((2, 2)), staff[2][0])
        >>> abjad.f(staff)
        \new Staff
        {
            \time 2/2
            c'2
            <d' f'>2
            \times 2/3 {
                \time 2/2
                g2
                a2
                b2
            }
        }
        >>> staff = auxjad.remove_repeated_time_signatures(staff)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 2/2
            c'2
            <d' f'>2
            \times 2/3 {
                g2
                a2
                b2
            }
        }
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("'container' must be 'abjad.Container' or child class")

    measures = [measure for measure in
                abjad.select(container).leaves().group_by_measure()]

    previous_time_signature = None
    for measure in measures:
        time_signature = abjad.inspect(measure[0]).indicator(
            abjad.TimeSignature
        )
        if time_signature == previous_time_signature:
            abjad.detach(abjad.TimeSignature, measure[0])
        if time_signature:
            previous_time_signature = time_signature

    return container



import random


class CartographyContainer():
    r"""A container used to store, manipulate, and select objects using a
    decaying weighted function.

    ..  container:: example

        The container should be initialised with a list of objects. The
        contents of the list can be absolutely anything.

        >>> container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
        >>> container.contents
        [0, 1, 2, 3, 4]

        The default decay rate is 0.75; that is, the weight of any given
        elements is the weight of the previous one multiplied by 0.75. The
        weights are associated with the index position, not the elements
        themselves.

        >>> container.weights
        [1.0, 0.75, 0.5625, 0.421875, 0.31640625]

        Applying the ``len()`` function to the container will give the length
        of the container.

        >>> len(container)
        5

        Calling the container will output one of its elements, selected
        according to its weight function.

        >>> result = ''
        >>> for _ in range(30):
        ...     result += str(container())
        >>> result
        203001402200011111101400310140

    ..  container:: example

        Calling the container with the optional keyword argument ``no_repeat``
        set to ``True`` will forbid immediate repetitions among consecutive
        calls.

        >>> container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
        >>> result = ''
        >>> for _ in range(30):
        ...     result += str(container(no_repeat=True))
        >>> result
        210421021020304024230120241202

    ..  container:: example

        The keyword argument ``decay_rate`` can be used to set a different
        decay rate when creating a container.

        >>> container = auxjad.CartographyContainer([0, 1, 2, 3, 4],
        ...                                         decay_rate=0.5,
        ...                                         )
        >>> cartography_container.weights
        [1.0, 0.5, 0.25, 0.125, 0.0625]

        The decay rate can also be set after the creation of a container,
        using the method ``set_decay_rate()``.

        >>> container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
        >>> container.set_decay_rate(0.2)
        >>> container.weights
        [1.0, 0.2, 0.04000000000000001, 0.008000000000000002,
        0.0016000000000000003]
        >>> result = ''
        >>> for _ in range(30):
        ...     result += str(container())
        >>> result
        '000001002100000201001030000100'

    ..  container:: example

        Appending is a type of container transformation. It discards the first
        element of the container, shifts all others leftwards, and then appends
        the new element to the rightmost index.

        >>> container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
        >>> container.contents
        [0, 1, 2, 3, 4]
        >>> container.append(5)
        >>> container.contents
        [1, 2, 3, 4, 5]
        >>> container.append(42)
        >>> container.contents
        [2, 3, 4, 5, 42]

    ..  container:: example

        The method ``append_keeping_n()`` is similar to ``append()``, but it
        keeps the first n indeces untouched. It thus discards the n+1-th
        element, shifts all the next ones lefwards and then appends the new
        element at the end of the container.

        >>> container = auxjad.CartographyContainer([10, 7, 14, 31, 98])
        >>> container.contents
        [10, 7, 14, 31, 98]
        >>> container.append_keeping_n(100, 2)
        >>> container.contents
        [10, 7, 31, 98, 100]

    ..  container:: example

        Prepending is another type of container transformation. It discards the
        last element of the container, shifts all others rightwards, and then
        prepends the new element to the leftmost index.

        >>> container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
        >>> container.contents
        [0, 1, 2, 3, 4]
        >>> container.prepend(-1)
        >>> container.contents
        [-1, 0, 1, 2, 3]
        >>> container.prepend(71)
        >>> container.contents
        [71, -1, 0, 1, 2]

    ..  container:: example

        Rotation is another type of container transformation. It rotates all
        elements rightwards, while moving the rightmost element into the
        leftmost index. It can take the optional keyword argument anticlockwise
        which if set to ``True`` will rotate in the opposite direction.

        >>> container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
        >>> container.contents
        [0, 1, 2, 3, 4]
        >>> container.rotate()
        >>> container.contents
        [1, 2, 3, 4, 0]
        >>> container.rotate(anticlockwise=True)
        >>> container.contents
        [0, 1, 2, 3, 4]
        >>> container.rotate(anticlockwise=True)
        >>> container.contents
        [1, 2, 3, 4, 0]

    ..  container:: example

        It is also possible to mirror two elements around a pivot at the centre
        of the container; given an element (selected by its index), this
        operation will locate and swap it for its complementary element. The
        complementary element is defined as that one which is at a same
        distance from the centre pivot but in the opposite direction.

        >>> container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
        >>> container.contents
        [0, 1, 2, 3, 4]
        >>> container.mirror(0)
        >>> container.contents
        [4, 1, 2, 3, 0]
        >>> container.mirror(0)
        >>> container.contents
        [0, 1, 2, 3, 4]
        >>> container.mirror(3)
        >>> container.contents
        [0, 3, 2, 1, 4]
        >>> container.mirror(2)
        >>> container.contents
        [0, 3, 2, 1, 4]

    ..  container:: example

        To mirror a random pair of complementary elements, use the
        ``mirror_random()`` method. In case of a container with an odd number
        of elements, this method will never pick an element at the pivot point
        since the operation would not change the contents.

        >>> container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
        >>> container.contents
        [0, 1, 2, 3, 4]
        >>> container.mirror_random()
        >>> container.contents
        [4, 1, 2, 3, 0]
        >>> container.mirror_random()
        >>> container.contents
        [4, 3, 2, 1, 0]
        >>> container.mirror_random()
        >>> container.contents
        [4, 1, 2, 3, 0]

    ..  container:: example

        The method ``randomise()`` will randomise the position of the elements
        of a container.

        >>> container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
        >>> container.contents
        [0, 1, 2, 3, 4]
        >>> container.randomise()
        >>> container.contents
        [1, 4, 3, 0, 2]

    ..  container:: example

        The contents of a container can also be altered after it has been
        initialised using the ``set_container()`` method. The length of the
        container can change too.

        >>> container = auxjad.CartographyContainer([0, 1, 2, 3, 4],
        ...                                         decay_rate=0.5,
        ...                                         )
        >>> len(container)
        5
        >>> container.weights
        [1.0, 0.5, 0.25, 0.125, 0.0625]
        >>> container.set_container([10, 7, 14, 31, 98, 47, 32])
        >>> container.contents
        [10, 7, 14, 31, 98, 47, 32]
        >>> len(container)
        7
        >>> container.weights
        [1.0, 0.5, 0.25, 0.125, 0.0625, 0.03125, 0.015625]

    ..  container:: example

        To method ``replace_element()`` replaces a specific element at a
        specified index.

        >>> container = auxjad.CartographyContainer([10, 7, 14, 31, 98])
        >>> container.contents
        [10, 7, 14, 31, 98]
        >>> container.replace_element(100, 2)
        >>> container.contents
        [10, 7, 100, 31, 98]

    ..  container:: example

        The attribute previous_index stores the previously selected index. It
        can be used with the ``get_element()`` method in order to retrieve the
        last value output by the object.

        >>> container = auxjad.CartographyContainer([10, 7, 14, 31, 98])
        >>> container()
        31
        >>> previous_index = container.previous_index
        >>> previous_index
        3
        >>> container.get_element(previous_index)
        31

    ..  container:: example

        Individual elements are also accessible via indeces of the content
        attribute. When accessed in this manner, they can also be sliced like a
        regular list.

        >>> container = auxjad.CartographyContainer([10, 7, 14, 31, 98])
        >>> container.contents[2]
        14
        >>> container.contents[1:4]
        [7, 14, 31]
    """

    def __init__(self,
                 container: list,
                 *,
                 decay_rate: float = 0.75,
                 ):
        if not isinstance(container, list):
            raise TypeError("'container' must be 'list'")
        if not isinstance(decay_rate, float):
            raise TypeError("'decay_rate' must be 'float'")
        if decay_rate <= 0.0 or decay_rate > 1.0:
            raise ValueError("'decay_rate' must be larger than 0.0 and "
                             "less than or equal to 1.0")
        self.contents = container[:]
        self.previous_index = None
        self._decay_rate = decay_rate
        self._generate_weights()

    def __call__(self, no_repeat=False):
        if not no_repeat:
            new_index = random.choices(
                [n for n in range(self.__len__())],
                weights=self.weights,
                )[0]
        else:
            new_index = self.previous_index
            while (new_index == self.previous_index):
                new_index = random.choices(
                    [n for n in range(self.__len__())],
                    weights=self.weights,
                    )[0]
        self.previous_index = new_index
        return self.contents[self.previous_index]

    def __len__(self):
        return len(self.contents)

    def append(self, new_element):
        self.contents = self.contents[1:] + [new_element]

    def append_keeping_n(self, new_element, n: int):
        self.contents = self.contents[:n] \
                      + self.contents[n+1:] \
                      + [new_element]

    def prepend(self, new_element):
        self.contents = [new_element] + self.contents[:-1]

    def rotate(self, anticlockwise=False):
        if not anticlockwise:
            self.contents = self.contents[1:] + self.contents[:1]
        else:
            self.contents = self.contents[-1:] + self.contents[:-1]

    def mirror(self, index: int):
        aux = self.contents[index]
        self.contents[index] = self.contents[-1 - index]
        self.contents[-1 - index] = aux

    def mirror_random(self):
        max_index = self.__len__() // 2 - 1
        self.mirror(random.randint(0, max_index))

    def randomise(self):
        random.shuffle(self.contents)

    def get_element(self, index: int):
        return self.contents[index]

    def replace_element(self, new_element, index: int):
        self.contents = self.contents[:index] \
                      + [new_element] \
                      + self.contents[index+1:]

    def set_container(self, new_container: list):
        if not isinstance(new_container, list):
            raise TypeError("'new_container' must be 'list")
        self.contents = new_container[:]
        self._generate_weights()

    def set_decay_rate(self, new_decay_rate: float):
        if not isinstance(new_decay_rate, float):
            raise TypeError("'new_decay_rate' must be float")
        if new_decay_rate <= 0.0 or new_decay_rate > 1.0:
            raise ValueError("'new_decay_rate' must be larger than 0.0 and "
                             "less than or equal to 1.0")
        self._decay_rate = new_decay_rate
        self._generate_weights()

    def _generate_weights(self):
        self.weights = []
        for n in range(self.__len__()):
            self.weights.append(self._decay_rate ** n)



import abjad


def are_leaves_tieable(leaf1: abjad.Leaf,
                     leaf2: abjad.Leaf,
                     ) -> bool:
    r"""A comparator function returning ``True`` when two leaves have identical
    pitches and thus can be tied, otherwise returning ``False``.

    ..  container:: example

        When the pitches in both leaves are identical, this function returns
        ``True``:

        >>> Leaf1 = abjad.Note(r"c'4")
        >>> Leaf2 = abjad.Note(r"c'4")
        >>> auxjad.are_leaves_tieable(Leaf1, Leaf2)
        True

    ..  container:: example

        Durations do not affect the comparison.

        >>> Leaf1 = abjad.Note(r"c'2.")
        >>> Leaf2 = abjad.Note(r"c'16")
        >>> Leaf3 = abjad.Note(r"f'''16")
        >>> auxjad.are_leaves_tieable(Leaf1, Leaf2)
        True
        >>> auxjad.are_leaves_tieable(Leaf1, Leaf3)
        False
        >>> auxjad.are_leaves_tieable(Leaf2, Leaf3)
        False

    ..  container:: example

        Handles chords as well as pitches.

        >>> chord1 = abjad.Chord(r"<c' e' g'>4")
        >>> chord2 = abjad.Chord(r"<c' e' g'>16")
        >>> chord3 = abjad.Chord(r"<f''' fs'''>16")
        >>> auxjad.are_leaves_tieable(chord1, chord2)
        True
        >>> auxjad.are_leaves_tieable(chord1, chord3)
        False
        >>> auxjad.are_leaves_tieable(chord2, chord3)
        False

    ..  container:: example

        Leaves can also be part of containers.

        >>> container = abjad.Container(r"r4 <c' e'>4 <c' e'>2")
        >>> auxjad.are_leaves_tieable(container[1], container[2])
        True

    ..  container:: example

        If rests are input, the return value is ``False``.

        >>> container = abjad.Container(r"r4 g'4 r2")
        >>> auxjad.are_leaves_tieable(container[0], container[2])
        False
    """
    if not isinstance(leaf1, abjad.Leaf):
        raise TypeError("'leaf1' must be 'abjad.Leaf' or child class")
    if not isinstance(leaf2, abjad.Leaf):
        raise TypeError("'leaf2' must be 'abjad.Leaf' or child class")

    if type(leaf1) != type(leaf2):
        return False
    if type(leaf1) == abjad.Rest:
        return False
    if type(leaf1) == abjad.MultimeasureRest:
        return False
    if type(leaf1) == abjad.Note:
        if leaf1.written_pitch != leaf2.written_pitch:
            return False
    if type(leaf1) == abjad.Chord:
        if leaf1.written_pitches != leaf2.written_pitches:
            return False
    if type(abjad.inspect(leaf1).before_grace_container()) != \
            type(abjad.inspect(leaf2).before_grace_container()):
        return False
    return True



import abjad
import pytest
import auxjad


def test_sync_containers_01():
    container1 = abjad.Container(r"\time 4/4 g'2.")
    container2 = abjad.Container(r"\time 4/4 c'1")
    auxjad.sync_containers(container1, container2)
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            g'2.
            r4
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
        }
        ''')


def test_sync_containers_02():
    container1 = abjad.Container(r"\time 4/4 g'1 | f'4")
    container2 = abjad.Container(r"\time 4/4 c'1")
    auxjad.sync_containers(container1, container2)
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            g'1
            %%% \time 1/4 %%%
            f'4
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 1/4 %%%
            R1 * 1/4
        }
        ''')


def test_sync_containers_03():
    container1 = abjad.Container(r"\time 4/4 g'1 | f'4")
    container2 = abjad.Container(r"\time 4/4 c'1 | d'1")
    auxjad.sync_containers(container1, container2)
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            g'1
            f'4
            r2.
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            d'1
        }
        ''')


def test_sync_containers_04():
    container1 = abjad.Container(r"\time 4/4 g'1")
    container2 = abjad.Container(r"\time 4/4 c'1 | d'1")
    auxjad.sync_containers(container1, container2)
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            g'1
            R1
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            d'1
        }
        ''')


def test_sync_containers_05():
    container1 = abjad.Container(r"\time 4/4 g'1")
    container2 = abjad.Container(r"\time 4/4 c'1 | d'1")
    auxjad.sync_containers(container1,
                           container2,
                           use_multimeasure_rests=False,
                           )
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            g'1
            r1
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            d'1
        }
        ''')


def test_sync_containers_06():
    container1 = abjad.Container(r"\time 3/4 g'2.")
    container2 = abjad.Container(r"\time 3/4 c'2. | d'2.")
    auxjad.sync_containers(container1, container2)
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 3/4 %%%
            g'2.
            R1 * 3/4
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 3/4 %%%
            c'2.
            d'2.
        }
        ''')


def test_sync_containers_07():
    container1 = abjad.Container(r"\time 4/4 c'1 | g'4")
    container2 = abjad.Container(r"\time 4/4 c'1 | g'2")
    container3 = abjad.Container(r"\time 4/4 c'1 | g'2.")
    container4 = abjad.Container(r"\time 4/4 c'1")
    auxjad.sync_containers(container1, container2, container3, container4)
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration() \
        == abjad.inspect(container3).duration() \
        == abjad.inspect(container4).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            g'4
            r2
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            g'2
            r4
        }
        ''')
    assert format(container3) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            g'2.
        }
        ''')
    assert format(container4) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            R1 * 3/4
        }
        ''')


def test_sync_containers_08():
    container1 = abjad.Container(r"\time 4/4 c'1 | g'4")
    container2 = abjad.Container(r"\time 4/4 c'1 | g'2")
    container3 = abjad.Container(r"\time 4/4 c'1 | g'2.")
    container4 = abjad.Container(r"\time 4/4 c'1")
    auxjad.sync_containers(container1,
                           container2,
                           container3,
                           container4,
                           use_multimeasure_rests=False,
                           )
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration() \
        == abjad.inspect(container3).duration() \
        == abjad.inspect(container4).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            g'4
            r2
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            g'2
            r4
        }
        ''')
    assert format(container3) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            g'2.
        }
        ''')
    assert format(container4) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            r2.
        }
        ''')


def test_sync_containers_09():
    container1 = abjad.Container(r"\time 4/4 c'1 | g'4")
    container2 = abjad.Container(r"\time 4/4 c'1 | g'2")
    container3 = abjad.Container(r"\time 4/4 c'1 | g'2.")
    container4 = abjad.Container(r"\time 4/4 c'1")
    auxjad.sync_containers(container1,
                           container2,
                           container3,
                           container4,
                           adjust_last_time_signature=False,
                           )
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration() \
        == abjad.inspect(container3).duration() \
        == abjad.inspect(container4).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            g'4
            r2
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            g'2
            r4
        }
        ''')
    assert format(container3) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            g'2.
        }
        ''')
    assert format(container4) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'1
            r2.
        }
        ''')


def test_sync_containers_10():
    container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4")
    container2 = abjad.Container(r"\time 3/4 a2. \time 4/4 c'4")
    container3 = abjad.Container(r"\time 5/4 g''1 ~ g''4")
    container4 = abjad.Container(r"\time 6/8 c'2")
    auxjad.sync_containers(container1, container2, container3, container4)
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration() \
        == abjad.inspect(container3).duration() \
        == abjad.inspect(container4).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'4
            d'4
            e'4
            f'4
            %%% \time 1/4 %%%
            R1 * 1/4
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 3/4 %%%
            a2.
            %%% \time 2/4 %%%
            c'4
            r4
        }
        ''')
    assert format(container3) == abjad.String.normalize(
        r'''
        {
            %%% \time 5/4 %%%
            g''1
            ~
            g''4
        }
        ''')
    assert format(container4) == abjad.String.normalize(
        r'''
        {
            %%% \time 6/8 %%%
            c'2
            r4
            %%% \time 2/4 %%%
            R1 * 1/2
        }
        ''')


def test_sync_containers_11():
    container1 = abjad.Container(r"\time 4/4 g'1 | f'4")
    container2 = abjad.Container(r"\time 5/4 c'1 | \time 4/4 d'4")
    with pytest.raises(ValueError):
        assert auxjad.sync_containers(container1, container2)


def test_sync_containers_12():
    container1 = abjad.Staff(r"\time 4/4 g'2.")
    container2 = abjad.Staff(r"\time 4/4 c'1")
    auxjad.sync_containers(container1, container2)
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 4/4
            g'2.
            r4
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 4/4
            c'1
        }
        ''')


def test_sync_containers_13():
    container1 = abjad.Container(r"\time 3/4 g'2.")
    container2 = abjad.Container(r"\time 3/4 c'2.")
    auxjad.sync_containers(container1, container2)
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 3/4 %%%
            g'2.
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 3/4 %%%
            c'2.
        }
        ''')


def test_sync_containers_14():
    container1 = abjad.Container(r"\time 5/4 g'1~g'4 | R1 * 5/4")
    container2 = abjad.Container(r"\time 5/4 c'2.")
    auxjad.sync_containers(container1, container2)
    assert abjad.inspect(container1).duration() \
        == abjad.inspect(container2).duration()
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 5/4 %%%
            g'1
            ~
            g'4
            R1 * 5/4
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 5/4 %%%
            c'2.
            r2
            R1 * 5/4
        }
        ''')
